import numpy as np

__all__ = [
    "Inf",
    "Infinity",
    "NAN",
    "NINF",
    "NZERO",
    "NaN",
    "PINF",
    "PZERO",
    "e",
    "euler_gamma",
    "inf",
    "infty",
    "nan",
    "newaxis",
    "pi",
]

Inf = np.Inf
Infinity = np.Infinity
NAN = np.NAN
NINF = np.NINF
NZERO = np.NZERO
NaN = np.NaN
PINF = np.PINF
PZERO = np.PZERO
e = np.e
euler_gamma = np.euler_gamma
inf = np.inf
infty = np.infty
nan = np.nan
newaxis = np.newaxis
pi = float(np.pi)
